/*
 * imgmirc.dll is a plug-in for the mIRC application
 * Copyright (C) 2003-2010 Ivan Zvolsky
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library (lesser.txt)
 */

#ifndef _IMGSTORAGE_
#define _IMGSTORAGE_

#define MAXREAD 32768

#include "common_storage.h"

class _img_storage;
class _image_item;

//--------------------------------------
// Here goes image item class definition
//
class _image_item
{
    friend _img_storage;
    friend _storage<_image_item>;


    BITMAPINFOHEADER    bi;
    WORD                offset;
    HDC                 memDC;
    HBITMAP             memBmp;

    _image_item *next;

    BOOL ReadDIB (void);

    union _data
    {
        LPBITMAPINFO lpbi;
        unsigned char *raw_data;
    } data;

public:
    char *imgname;

    _image_item() : imgname(NULL) { data.raw_data = NULL; }
    ~_image_item()
    {
        if (imgname)
            delete imgname;
        if (data.raw_data)
            delete data.raw_data;
        DeleteObject(memBmp);
        DeleteObject(memDC);
    }
    _image_item(LPCTSTR _imgname)
    {
        imgname = new char[strlen(_imgname) + 1];
        sprintf(imgname, "%s\x0", _imgname);
    }

    void _image_item::TranspDrawImage(HDC hdc, POINT *point, RECT *rect)
    {
        TransparentBlt(hdc, point->x, point->y,
                       rect->right, rect->bottom,
                       memDC, rect->left, rect->top,
                       rect->right, rect->bottom, RGB(255, 255, 255));
    }

    void _image_item::TranspStretchImage(HDC hdc, POINT *point, RECT *rect, int dest_x)
    {
        int x = point->x +  rect->right / 2 - (rect->right * dest_x / rect->bottom) / 2;
        int y = point->y + rect->bottom / 2 - dest_x / 2;
        TransparentBlt(hdc, x, y,
                       rect->right * dest_x / rect->bottom, dest_x,
                       memDC, rect->left, rect->top,
                       rect->right, rect->bottom, RGB(255, 255, 255));
    }

    void _image_item::JustDrawImage(HDC hdc, POINT *point, RECT *rect)
    {
        SetDIBitsToDevice(hdc,
                          point->x, point->y,
                          rect->right, rect->bottom,
                          rect->left,
                          data.lpbi->bmiHeader.biHeight - rect->top-rect->bottom,
                          0, data.lpbi->bmiHeader.biHeight, data.raw_data + offset,
                          data.lpbi, DIB_RGB_COLORS);
    }

    void _image_item::JustStretchImage(HDC hdc, POINT *point, RECT *rect, int dest_x)
    {
        int x = point->x +  rect->right / 2 - (rect->right * dest_x / rect->bottom) / 2;
        int y = point->y + rect->bottom / 2 - dest_x / 2;
        TransparentBlt(hdc, x, y,
                       rect->right * dest_x / rect->bottom, dest_x,
                       memDC, rect->left, rect->top,
                       rect->right, rect->bottom, RGB(255, 255, 255));
    }
};

//-----------------------------------------
// Here goes image storage class definition
//
class _img_storage:_storage<_image_item>
{
    HBITMAP hBmp;
    HDC bmpDC;
public:
    _img_storage();
    ~_img_storage();
    _image_item * PutItem(LPCTSTR name);
    void DrawAll(HDC hDC);
};

extern _img_storage imagestorage;

#endif