/*
 * imgmirc.dll is a plug-in for the mIRC application
 * Copyright (C) 2003-2010 Ivan Zvolsky
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library (lesser.txt)
 */
#include "stdafx.h"
#include "my_funcs.h"
#include "flags.h"
#include "pcre.h"

#include "replstorage.h"
#include "highlights.h"
#include "imgstorage.h"
#include "stdlib.h"
#include "textout.h"

struct _func_tag func_tags[funcs_number];

void return_functions()
{
    SIZE_T written;
    DWORD op;
    for (int i=0; i<funcs_number; i++)
        if (func_tags[i].isd)
        {
            VirtualProtect((void*)(func_tags[i].isd), 4, PAGE_READWRITE, &op);
            WriteProcessMemory(GetCurrentProcess(), (void*)(func_tags[i].isd),
                               (void*)&(func_tags[i].old_func), 4, &written);
            VirtualProtect((void*)(func_tags[i].isd), 4, op, &op);

            if (written != 4)
            {
                conv_echo("imgmirc: Unable to rewrite address", 4);
                return;
            } else
                verb_conv_echo(4, "imgmirc: %x address returned", func_tags[i].old_func);
        }
}

void do_intercept(void)
{
    for (int i=0; i<funcs_number; i++)
        func_tags[i].isd = NULL;

    if (_PCRE_HIGHLIGHTS_)
    {
        if (mirc_major_ver >= 7)
        {
            verb_conv_echo("intercepting recv (wsock32.dll)....", 7);
            int res = intercept_function("WSOCK32.dll", "recv", func_recv, (DWORD)&my_recv);

            if (res) {
                verb_conv_echo("intercepting recv (ws2_32.dll)....", 7);
                intercept_function("ws2_32.dll", "recv", func_recv, (DWORD)&my_recv);
            }
        } else
        {
            verb_conv_echo("intercepting recv....", 7);
            intercept_function("WSOCK32.dll", "recv", func_recv, (DWORD)&my_recv);
        }
    }
    if (_PCRE_HIGHLIGHTS_ || _SMILES_)
    {
        if (mirc_major_ver >= 7)
        {
            verb_conv_echo("imgmirc: intercepting ExtTextOutW....", 7);
            intercept_function("GDI32.DLL", "ExtTextOutW", func_textout_wchar,
                                _TRANSPARENT_ ? (DWORD)&my_TranspExtTextOutW : (DWORD)&my_SimpleExtTextOutW);

            // starting from 6.34 (?) we need to intercept DrawText() - they changed the way text is drawn
            if ((mirc_major_ver == 6 && mirc_minor_ver >= 34) || mirc_major_ver >= 7) {
                    verb_conv_echo("intercepting DrawTextW....", 7);
                    intercept_function("USER32.dll", "DrawTextW", func_drawtext_wchar,
                                       _TRANSPARENT_ ? (DWORD)&my_TransparentDrawTextW : (DWORD)&my_SimpleDrawTextW);
            }
        } else
        {
            verb_conv_echo("imgmirc: intercepting ExtTextOutA....", 7);
            intercept_function("GDI32.DLL", "ExtTextOutA", func_textout_char,
                                _TRANSPARENT_ ? (DWORD)&my_TranspExtTextOutA : (DWORD)&my_SimpleExtTextOutA);

            // starting from 6.34 (?) we need to intercept DrawText() - they changed the way text is drawn
            if ((mirc_major_ver == 6 && mirc_minor_ver >= 34) || mirc_major_ver >= 7) {
                    verb_conv_echo("intercepting DrawTextA....", 7);
                    intercept_function("USER32.dll", "DrawTextA", func_drawtext_char,
                                       _TRANSPARENT_ ? (DWORD)&my_TransparentDrawTextA : (DWORD)&my_SimpleDrawTextA);
            }
        }
    }
}

int intercept_function(LPCTSTR module_name,
                       LPCTSTR func_name,
                       _func_desc num,
                       DWORD new_func)
{
    func_tags [num].isd = 0;

    BYTE *pimage = (BYTE*)GetModuleHandle(NULL);

    IMAGE_DOS_HEADER *idh;
    IMAGE_OPTIONAL_HEADER *ioh;
    IMAGE_SECTION_HEADER *ish;
    IMAGE_IMPORT_DESCRIPTOR *iid;
    DWORD *isd;  //image_thunk_data dword

    idh = (IMAGE_DOS_HEADER*)pimage;
    ioh = (IMAGE_OPTIONAL_HEADER*)(pimage + idh->e_lfanew + 4 + sizeof(IMAGE_FILE_HEADER));
    ish = (IMAGE_SECTION_HEADER*)((BYTE*)ioh + sizeof(IMAGE_OPTIONAL_HEADER));

    if (idh->e_magic != 0x5A4D)
    {
        conv_echo("imgmirc: Not exe hdr", 4);
        return 1;
    }

    int i;
    for(i=0; i<16; i++)
        if (strcmp((char*)((ish + i)->Name), ".idata") == 0)
            break;

    if (i == 16)
    {
        verb_conv_echo("imgmirc: Unable to find .idata section", 2);
        iid = (IMAGE_IMPORT_DESCRIPTOR*)(pimage + ioh->DataDirectory[IMAGE_DIRECTORY_ENTRY_IMPORT].VirtualAddress);
    } else
        iid = (IMAGE_IMPORT_DESCRIPTOR*)(pimage + (ish + i)->VirtualAddress );

    func_tags[num].old_func = (DWORD)GetProcAddress(GetModuleHandle(module_name), func_name);
    if (func_tags[num].old_func == 0)
    {
        conv_echo(4, "imgmirc: Can`t get address of %s function", func_name);
        return 1;
    } else
        verb_conv_echo(4, "imgmirc: address of %s is %x", func_name, func_tags[num].old_func);

    while(iid->Name)
    {
        if (_stricmp((char*)(pimage + iid->Name), module_name) == 0 )
            break;
        iid++;
    }

    isd = (DWORD*)(pimage + iid->FirstThunk);
    while (*isd != func_tags[num].old_func && *isd != 0)
        isd++;

    if(*isd == 0)
    {
        conv_echo("imgmirc: addr_func not found in .idata", 4);
        return 1;
    }

    DWORD buf = new_func;
    DWORD op;
    SIZE_T written;

    VirtualProtect((void*)(isd), 4, PAGE_READWRITE, &op);
    WriteProcessMemory(GetCurrentProcess(), (void*)(isd), (void*)&buf, 4, &written);
    VirtualProtect((void*)(isd), 4, op, &op);

    if (written != 4)
    {
        conv_echo("imgmirc: Unable to rewrite address", 4);
        return 1;
    }

    func_tags [num].new_func = new_func;
    func_tags [num].isd = isd;

    return 0; //success
}

int WINAPI my_recv(SOCKET s, char FAR *buf, int len, int flags)
{
    int ret;
    ret = CALL_RECV(s, buf, len, flags);

    std::vector<_repl_item *>::iterator r, r_end;
    std::vector<_hl_item *>::iterator i, i_end;

    int gl_pos;
    char buffer[8192]; // FIXME

    if (ret > 0)
    {
        // performing highlighting
        BOOL flag;
        gl_pos = 0;
        int b_pos = 0;

        while (hl_privmsg->hl_exec(buf, ret, gl_pos) > 0)
        {
            flag = FALSE;

            i_end = hl_list.end();
            for (i=hl_list.begin(); i<i_end && !flag; i++)
                if ((*i)->hl_exec(buf + hl_privmsg->ovector[1], ret - hl_privmsg->ovector[1], 0) > 0)
                {
                    memcpy(buffer + b_pos, buf + gl_pos, hl_privmsg->ovector[1] - gl_pos + (*i)->ovector[1]);
                    b_pos += hl_privmsg->ovector[1] - gl_pos + (*i)->ovector[1];

                    buffer[b_pos++] = ' ';
                    memcpy(buffer + b_pos, hl_mainstr, HL_STUB_LENGTH);
                    b_pos += HL_STUB_LENGTH;

                    flag = TRUE;
                    gl_pos = hl_privmsg->ovector[1] + (*i)->ovector[1];
                }

                if (!flag)
                {
                    memcpy(buffer + b_pos, buf + gl_pos, ret - gl_pos);
                    b_pos += ret - gl_pos;
                    gl_pos = ret;
                }
        }

        if (gl_pos < ret)
        {
            memcpy(buffer + b_pos, buf + gl_pos, ret - gl_pos);
            b_pos += ret - gl_pos;
            memcpy(buf, buffer, b_pos);
            ret = b_pos;
        }
    }
    return ret;
}

int WINAPI my_TransparentDrawTextA(HDC hDC,          // handle to DC
                                   LPCTSTR lpchText, // text to draw
                                   int nCount,       // text length
                                   LPRECT lpRect,    // formatting dimensions
                                   UINT uFormat      // text-drawing options
                                   )
{
    if (uFormat & DT_CALCRECT
        || uFormat == 37 /* dirty hack for multi-tab control like "address book" in 6.34 */)
        return CALL_DRAWTEXT_A(hDC, lpchText, nCount, lpRect, uFormat);
    else {
        my_TranspExtTextOutA(hDC, lpRect->left, lpRect->top, 0, lpRect, lpchText, nCount, NULL);
        return lpRect->bottom - lpRect->top;
    }
}

int WINAPI my_SimpleDrawTextA(HDC hDC,          // handle to DC
                              LPCTSTR lpchText, // text to draw
                              int nCount,       // text length
                              LPRECT lpRect,    // formatting dimensions
                              UINT uFormat      // text-drawing options
                              )
{
    if (uFormat & DT_CALCRECT
        || uFormat == 37 /* dirty hack for multi-tab control like "address book" in 6.34 */)
        return CALL_DRAWTEXT_A(hDC, lpchText, nCount, lpRect, uFormat);
    else {
        my_SimpleExtTextOutA(hDC, lpRect->left, lpRect->top, 0, lpRect, lpchText, nCount, NULL);
        return lpRect->bottom - lpRect->top;
    }
}

int WINAPI my_TransparentDrawTextW(HDC hDC,          // handle to DC
                                   LPCWSTR lpchText, // text to draw
                                   int nCount,       // text length
                                   LPRECT lpRect,    // formatting dimensions
                                   UINT uFormat      // text-drawing options
                                  )
{
    if (uFormat & DT_CALCRECT
        || uFormat == 37 /* dirty hack for multi-tab control like "address book" in 6.34 */)
        return CALL_DRAWTEXT_W(hDC, lpchText, nCount, lpRect, uFormat);
    else {
        my_TranspExtTextOutW(hDC, lpRect->left, lpRect->top, 0, lpRect, lpchText, nCount, NULL);
        return lpRect->bottom - lpRect->top;
    }
}

int WINAPI my_SimpleDrawTextW(HDC hDC,          // handle to DC
                              LPCWSTR lpchText, // text to draw
                              int nCount,       // text length
                              LPRECT lpRect,    // formatting dimensions
                              UINT uFormat      // text-drawing options
                             )
{
    if (uFormat & DT_CALCRECT
        || uFormat == 37 /* dirty hack for multi-tab control like "address book" in 6.34 */)
        return CALL_DRAWTEXT_W(hDC, lpchText, nCount, lpRect, uFormat);
    else {
        my_SimpleExtTextOutW(hDC, lpRect->left, lpRect->top, 0, lpRect, lpchText, nCount, NULL);
        return lpRect->bottom - lpRect->top;
    }
}
