/*
 * imgmirc.dll is a plug-in for the mIRC application
 * Copyright (C) 2003-2010 Ivan Zvolsky
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library (lesser.txt)
 */
#include "stdafx.h"
#include "replstorage.h"
#include "flags.h"
#include "convthread.h"
#include "exceptions.h"

_repl_item::_repl_item(LPCTSTR hl_str, LPCTSTR img_str, RECT *prect_image)
{
    img = NULL;
    hl = NULL;
    if (!(img = imagestorage.PutItem(img_str)))
        throw file_error;
    hl = new _hl_item(hl_str);
    if (prect_image)
        rect_image = *prect_image;
}

_repl_item::~_repl_item()
{
    if (hl)
        delete hl;
}

void LoadReplacementsList()
{
    repl_list.reserve(32);
    if (_TRANSPARENT_)
        text_list.reserve(32);

    char buffer[1024];

    FILE * stream;
    sprintf(buffer, "%simgmirc\\imgs\\replacements.txt", mircDir);
    if (!(stream = fopen(buffer, "rb")))
        throw replfile_error;

    const char *error;
    int erroffset;
    pcre *re = NULL;
    pcre_extra *pe = NULL;
    int *ovector = NULL;
    int np;
    re = pcre_compile("(.*)#([^#]+)#(\\d+)#(\\d+)#(\\d+)#(\\d+)",
                      0, &error, &erroffset, NULL);
    if (!re)
        return;
    pe = pcre_study(re, 0, &error);
    // get number of subpatterns
    np = pcre_info(re, NULL, NULL);
    ovector = (int*)malloc((np + 1) * 3 * sizeof(int));
    int ret;

    _repl_item *new_item;

    while (fscanf(stream, "%[^\n]s", buffer) != EOF)
    {
        if ((ret = pcre_exec(re, pe, buffer, strlen(buffer), 0, 0, ovector, (np + 1) * 3)) == 7)
        {
            buffer[ovector[3]] = 0;
            buffer[ovector[5]] = 0;
            try {
                RECT rect;
                sscanf(buffer + ovector[6], "%d#%d#%d#%d", &rect.left, &rect.top, &rect.right, &rect.bottom);
                new_item = new _repl_item(buffer + ovector[2], buffer + ovector[4], &rect);
                repl_list.push_back(new_item);
            }
            catch (int i)
            {
                switch (i)
                {
                    case compile_error:
                        verb_conv_echo(4, "pcre: can not compile pattern '%s'", buffer);
                        break;
                    case file_error:
                        verb_conv_echo(4, "imgmirc: can not open file '%s'", buffer + ovector[4]);
                        break;
                    default:
                        conv_echo(4, "imgmirc: %d exception catched", i);
                }
            }
        }
        else
            conv_echo(4, "igmirc: replacements.txt:'%s' string ignored!", buffer);

        fseek(stream, 1, SEEK_CUR);
    }

    if (re)
        pcre_free(re);
    if (pe)
        pcre_free(pe);
    if (ovector)
        free(ovector);

    fclose(stream);
}

std::vector<_repl_item *> repl_list;
std::vector<RECT> text_list;
