/*
 * imgmirc.dll is a plug-in for the mIRC application
 * Copyright (C) 2003-2010 Ivan Zvolsky
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library (lesser.txt)
 */
#include "stdafx.h"
#include "my_funcs.h"
#include "flags.h"
#include "pcre.h"

#include "replstorage.h"
#include "highlights.h"
#include "imgstorage.h"
#include "stdlib.h"
#include "textout.h"

BOOL my_rect_compare(const RECT & rect1, const RECT & rect2)
{
    return (rect1.top < rect2.top);
}

BOOL WINAPI my_SimpleExtTextOutA(HDC hdc,          // handle to DC
                                 int X,            // x-coordinate of reference point
                                 int Y,            // y-coordinate of reference point
                                 UINT fuOptions,   // text-output options
                                 const RECT* lprc, // optional dimensions
                                 LPCTSTR lpString, // string
                                 UINT cbCount,     // number of characters in string
                                 const int* lpDx   // array of spacing values
                                 )
{
    SIZE sz, sz1;
    std::vector<_repl_item *>::iterator i, i_end;
    std::vector<RECT>::iterator j, j_end;

    int p;
    BOOL ret;
    char * pos;
    HPEN f_pen;

    if (cbCount < 2)
        return CALL_TEXTOUT_A(hdc, X, Y, fuOptions, lprc, lpString, cbCount, lpDx);

    if (_PCRE_HIGHLIGHTS_)
        if (pos = (char*)strstr(lpString, hl_mainstr))
            memcpy(pos, HL_STUB_STRING, HL_STUB_LENGTH);

    ret = CALL_TEXTOUT_A(hdc, X, Y, fuOptions, lprc, lpString, cbCount, lpDx);

    if (_SMILES_)
    {
        f_pen = (HPEN)SelectObject(hdc, (HPEN)GetStockObject(NULL_PEN));
        i_end = repl_list.end();
        for (i=repl_list.begin(); i<i_end; i++)
        {
            p = 0;
            while ((*i)->hl->hl_exec((char *)lpString, cbCount, p) > 0)
            {
                GetTextExtentPoint32(hdc, lpString + (*i)->hl->ovector[2], (*i)->hl->ovector[3] - (*i)->hl->ovector[2], &sz1);
                GetTextExtentPoint32(hdc, lpString, (*i)->hl->ovector[2], &sz);

                POINT pnt = {X + sz.cx + (sz1.cx - (*i)->rect_image.right)  / 2,
                                 Y - 1 + (sz1.cy - (*i)->rect_image.bottom) / 2};
                Rectangle(hdc, X + sz.cx, Y, X + sz1.cx + sz.cx + 1, Y + sz1.cy + 1);

                if (sz1.cy > (*i)->rect_image.bottom)
                    (*i)->img->JustDrawImage(hdc, &pnt, &(*i)->rect_image);
                else
                    (*i)->img->JustStretchImage(hdc, &pnt, &(*i)->rect_image, sz1.cy);

                p = (*i)->hl->ovector[3];
            }
        }
        SelectObject(hdc, f_pen);
    }
    return ret;
}

BOOL WINAPI my_TranspExtTextOutA(HDC hdc,          // handle to DC
                                 int X,            // x-coordinate of reference point
                                 int Y,            // y-coordinate of reference point
                                 UINT fuOptions,   // text-output options
                                 const RECT* lprc, // optional dimensions
                                 LPCTSTR lpString, // string
                                 UINT cbCount,     // number of characters in string
                                 const int* lpDx   // array of spacing values
                                 )
{
    SIZE sz, sz1;
    std::vector<_repl_item *>::iterator i, i_end;
    std::vector<RECT>::iterator j, j_end;

    int p;
    BOOL ret = true;
    char * pos;
    RECT text_pnt;

    if (cbCount < 2)
        return CALL_TEXTOUT_A(hdc, X, Y, fuOptions, lprc, lpString, cbCount, lpDx);

    if (_PCRE_HIGHLIGHTS_)
        if (pos = (char *)strstr(lpString, hl_mainstr))
            memcpy(pos, HL_STUB_STRING, HL_STUB_LENGTH);

    if (_SMILES_)
    {
        i_end = repl_list.end();
        for (i=repl_list.begin(); i<i_end; i++)
        {
            p = 0;
            while ((*i)->hl->hl_exec((char *)lpString, cbCount, p) > 0)
            {
                GetTextExtentPoint32A(hdc, lpString + (*i)->hl->ovector[2], (*i)->hl->ovector[3] - (*i)->hl->ovector[2], &sz1);
                GetTextExtentPoint32A(hdc, lpString, (*i)->hl->ovector[2], &sz);

                text_pnt.left   = X + sz.cx;
                text_pnt.right  = X + sz.cx + sz1.cx;
                text_pnt.top    = (*i)->hl->ovector[2];
                text_pnt.bottom = (*i)->hl->ovector[3] - (*i)->hl->ovector[2];

                text_list.push_back(text_pnt);

                POINT pnt = {X + sz.cx + (sz1.cx - (*i)->rect_image.right)  / 2,
                                 Y - 1 + (sz1.cy - (*i)->rect_image.bottom) / 2};

                text_pnt.top = Y;
                text_pnt.bottom = Y + sz1.cy + 1;
                CALL_TEXTOUT_A(hdc, X, Y, fuOptions, &text_pnt, "", 0, lpDx);

                if (sz1.cy > (*i)->rect_image.bottom)
                    (*i)->img->TranspDrawImage(hdc, &pnt, &(*i)->rect_image);
                else
                    (*i)->img->TranspStretchImage(hdc, &pnt, &(*i)->rect_image, sz1.cy);

                p = (*i)->hl->ovector[3];
            }
        }

        //sort here
        j_end = text_list.end();
        if (j_end != text_list.begin())
        {
            std::sort(text_list.begin(), text_list.end(), my_rect_compare);
            j_end = text_list.end();

            j = text_list.begin();
            text_pnt.top = Y;
            text_pnt.bottom = Y + sz1.cy + 1;
            text_pnt.left = X;
            text_pnt.right = j->left;

            CALL_TEXTOUT_A(hdc, X, Y, fuOptions, &text_pnt,
                           lpString,
                           text_list.begin()->top, lpDx);

            j_end--;
            for (j=text_list.begin(); j<j_end; j++)
            {
                text_pnt.left = j->right;
                text_pnt.right = (j+1)->left;

                CALL_TEXTOUT_A(hdc, j->right, Y, fuOptions, &text_pnt,
                               lpString + j->top + j->bottom,
                               (j+1)->top - (j->top + j->bottom), lpDx);
            }

            GetTextExtentPoint32A(hdc, lpString, cbCount, &sz);

            text_pnt.left = j_end->right;
            text_pnt.right = X + sz.cx;

            CALL_TEXTOUT_A(hdc, j_end->right, Y, fuOptions, &text_pnt,
                           lpString + j_end->top + j_end->bottom,
                           cbCount - (j_end->top + j_end->bottom), lpDx);
        } else
            ret = CALL_TEXTOUT_A(hdc, X, Y, fuOptions, lprc, lpString, cbCount, lpDx);

        text_list.clear();
    } else
        ret = CALL_TEXTOUT_A(hdc, X, Y, fuOptions, lprc, lpString, cbCount, lpDx);

    return ret;
}

BOOL WINAPI my_SimpleExtTextOutW(HDC hdc,          // handle to DC
                                 int X,            // x-coordinate of reference point
                                 int Y,            // y-coordinate of reference point
                                 UINT fuOptions,   // text-output options
                                 const RECT* lprc, // optional dimensions
                                 LPCWSTR lpString, // string
                                 UINT cbCount,     // number of characters in string
                                 const int* lpDx   // array of spacing values
                                 )
{
    SIZE sz, sz1;
    std::vector<_repl_item *>::iterator i, i_end;
    std::vector<RECT>::iterator j, j_end;

    int p;
    BOOL ret;
    char * pos;
    HPEN f_pen;

    if (cbCount < 2)
        return CALL_TEXTOUT_W(hdc, X, Y, fuOptions, lprc, lpString, cbCount, lpDx);

    // Convert to a char*
    size_t origsize = wcslen(lpString) + 1;
    size_t convertedChars = 0;
    char nstring[8192];
    wcstombs_s(&convertedChars, nstring, origsize, lpString, _TRUNCATE);

    if (_PCRE_HIGHLIGHTS_)
        if (pos = (char *)strstr(nstring, hl_mainstr))
            memcpy((void*)(lpString + (pos - nstring)), LHL_STUB_STRING, 2 * HL_STUB_LENGTH);

    ret = CALL_TEXTOUT_W(hdc, X, Y, fuOptions, lprc, lpString, cbCount, lpDx);

    if (_SMILES_)
    {
        f_pen = (HPEN)SelectObject(hdc, (HPEN)GetStockObject(NULL_PEN));
        i_end = repl_list.end();
        for (i=repl_list.begin(); i<i_end; i++)
        {
            p = 0;
            while ((*i)->hl->hl_exec(nstring, cbCount, p) > 0)
            {
                GetTextExtentPoint32W(hdc, lpString + (*i)->hl->ovector[2], (*i)->hl->ovector[3] - (*i)->hl->ovector[2], &sz1);
                GetTextExtentPoint32W(hdc, lpString, (*i)->hl->ovector[2], &sz);

                POINT pnt = {X + sz.cx + (sz1.cx - (*i)->rect_image.right)  / 2,
                                 Y - 1 + (sz1.cy - (*i)->rect_image.bottom) / 2};
                Rectangle(hdc, X + sz.cx, Y, X + sz1.cx + sz.cx + 1, Y + sz1.cy + 1);

                if (sz1.cy > (*i)->rect_image.bottom)
                    (*i)->img->JustDrawImage(hdc, &pnt, &(*i)->rect_image);
                else
                    (*i)->img->JustStretchImage(hdc, &pnt, &(*i)->rect_image, sz1.cy);

                p = (*i)->hl->ovector[3];
            }
        }
        SelectObject(hdc, f_pen);
    }
    return ret;
}

BOOL WINAPI my_TranspExtTextOutW(HDC hdc,          // handle to DC
                                 int X,            // x-coordinate of reference point
                                 int Y,            // y-coordinate of reference point
                                 UINT fuOptions,   // text-output options
                                 const RECT* lprc, // optional dimensions
                                 LPCWSTR lpString, // string
                                 UINT cbCount,     // number of characters in string
                                 const int* lpDx   // array of spacing values
                                 )
{
    SIZE sz, sz1;
    std::vector<_repl_item *>::iterator i, i_end;
    std::vector<RECT>::iterator j, j_end;

    int p;
    BOOL ret = true;
    char * pos;
    RECT text_pnt;

    if (cbCount < 2)
        return CALL_TEXTOUT_W(hdc, X, Y, fuOptions, lprc, lpString, cbCount, lpDx);

    // Convert to a char*
    size_t origsize = wcslen(lpString) + 1;
    size_t convertedChars = 0;
    char nstring[8192];
    wcstombs_s(&convertedChars, nstring, origsize, lpString, _TRUNCATE);

    if (_PCRE_HIGHLIGHTS_)
        if (pos = (char *)strstr(nstring, hl_mainstr))
            memcpy((void*)(lpString + (pos - nstring)), LHL_STUB_STRING, 2 * HL_STUB_LENGTH);

    if (_SMILES_)
    {
        i_end = repl_list.end();
        for (i=repl_list.begin(); i<i_end; i++)
        {
            p = 0;
            while ((*i)->hl->hl_exec((char *)nstring, cbCount, p) > 0)
            {
                GetTextExtentPoint32W(hdc, lpString + (*i)->hl->ovector[2], (*i)->hl->ovector[3] - (*i)->hl->ovector[2], &sz1);
                GetTextExtentPoint32W(hdc, lpString, (*i)->hl->ovector[2], &sz);

                text_pnt.left   = X + sz.cx;
                text_pnt.right  = X + sz.cx + sz1.cx;
                text_pnt.top    = (*i)->hl->ovector[2];
                text_pnt.bottom = (*i)->hl->ovector[3] - (*i)->hl->ovector[2];

                text_list.push_back(text_pnt);

                POINT pnt = {X + sz.cx + (sz1.cx - (*i)->rect_image.right)  / 2,
                                 Y - 1 + (sz1.cy - (*i)->rect_image.bottom) / 2};

                text_pnt.top = Y;
                text_pnt.bottom = Y + sz1.cy + 1;
                CALL_TEXTOUT_W(hdc, X, Y, fuOptions, &text_pnt, L"", 0, lpDx);

                if (sz1.cy > (*i)->rect_image.bottom)
                    (*i)->img->TranspDrawImage(hdc, &pnt, &(*i)->rect_image);
                else
                    (*i)->img->TranspStretchImage(hdc, &pnt, &(*i)->rect_image, sz1.cy);

                p = (*i)->hl->ovector[3];
            }
        }

        //sort here
        j_end = text_list.end();
        if (j_end != text_list.begin())
        {
            std::sort(text_list.begin(), text_list.end(), my_rect_compare);
            j_end = text_list.end();

            j = text_list.begin();
            text_pnt.top = Y;
            text_pnt.bottom = Y + sz1.cy + 1;
            text_pnt.left = X;
            text_pnt.right = j->left;

            CALL_TEXTOUT_W(hdc, X, Y, fuOptions, &text_pnt,
                           lpString,
                           text_list.begin()->top, lpDx);

            j_end--;
            for (j=text_list.begin(); j<j_end; j++)
            {
                text_pnt.left = j->right;
                text_pnt.right = (j+1)->left;

                CALL_TEXTOUT_W(hdc, j->right, Y, fuOptions, &text_pnt,
                               lpString + j->top + j->bottom,
                               (j+1)->top - (j->top + j->bottom), lpDx);
            }

            GetTextExtentPoint32W(hdc, lpString, cbCount, &sz);

            text_pnt.left = j_end->right;
            text_pnt.right = X + sz.cx;

            CALL_TEXTOUT_W(hdc, j_end->right, Y, fuOptions, &text_pnt,
                           lpString + j_end->top + j_end->bottom,
                           cbCount - (j_end->top + j_end->bottom), lpDx);
        } else
            ret = CALL_TEXTOUT_W(hdc, X, Y, fuOptions, lprc, lpString, cbCount, lpDx);

        text_list.clear();
    } else
        ret = CALL_TEXTOUT_W(hdc, X, Y, fuOptions, lprc, lpString, cbCount, lpDx);

    return ret;
}
